# !/bin/bash
# Custom script to create SD Card image for Webinar with STM32MP1
# WARNING: Experimental, not working with other targets

# [Internal] Source utilities scripts 
if [ -r "$PLATFORM_LOC/scripts/utils.sh" ]; then
	source "$PLATFORM_LOC/scripts/utils.sh"
else
	source "$script_lib_dir/utils.sh"
fi

BUILDSTEP_SCRIPT_DIR=$(dirname $BUILDSTEP_SCRIPT_PATH)
source ${BUILDSTEP_SCRIPT_DIR}/sdcardBuilder.sh ${LAYOUT}

do_buildstep_prepend

# Scale factor to get valid rootfs image size
# Note: The image size has to be bigger than the folder size, it ensure no error,
# 		The user can manually reduce image size afterward
: ${ROOTFS_SCALE_FACTOR:=1.3}

# Search tsv file 

if [ -z ${TSVFILE} ] ; then
	if [ -f ${PLATFORM_LOC}/sdcard_layout_mp1_sw4l.tsv ] ; then
		LAYOUT="${PLATFORM_LOC}/sdcard_layout_mp1_sw4l.tsv"
	else
		if [ -f "sdcard_layout_mp1_sw4l.tsv" ] ; then
			LAYOUT="sdcard_layout_mp1_sw4l.tsv"
		else
			echo "Error cannot find sdcard_layout_mp1_sw4l.tsv"
			exit 1
		fi
	fi
else
	LAYOUT=${TSVFILE}
fi


# create temporary folder
mkdir -p temp
rm -rf temp/*
imagesDir=temp

# get boot folder and format it with the right format and provide ${BOOTFS}
imgRoot="${PROJECT_LOC}/output/boot"
cp ${imgRoot} ${imagesDir} -r #
bootPartOutput="bootPart"
tempBoot="${imagesDir}/boot" #
#mv extlinux ${tempBoot} #

mkdir -p ${tempBoot}/extlinux

echo -e "# Generic Distro Configuration file generated by System Workbench For Linux
menu title Select the boot mode
MENU BACKGROUND ../splash.bmp
TIMEOUT 20
DEFAULT ${machine,,}-m4-examples-sdcard
LABEL ${machine,,}-sdcard
	KERNEL /uImage
	FDT /${machine,,}.dtb
	APPEND root=/dev/mmcblk0p5 rootwait rw console=ttySTM0,115200
LABEL ${machine,,}-a7-examples-sdcard
	KERNEL /uImage
	FDT /${machine,,}-a7-examples.dtb
	APPEND root=/dev/mmcblk0p5 rootwait rw console=ttySTM0,115200
LABEL ${machine,,}-m4-examples-sdcard
	KERNEL /uImage
	FDT /${machine,,}-m4-examples.dtb
	APPEND root=/dev/mmcblk0p5 rootwait rw console=ttySTM0,115200" > ${tempBoot}/extlinux/extlinux.conf

# Get size of the boot part # ${PROJECT_LOC}/output/boot
defaultSize=$(du -ks ${tempBoot}  | sed 's/\t.*//')    #We know that the integer count will be follow by a \t and then some scrap
defaultSize=$(expr ${defaultSize}*${ROOTFS_SCALE_FACTOR} | bc)
imgSize=${defaultSize%%.*}


log_exec generate_rootfs ${tempBoot} vfat ${bootPartOutput} BOOT ${imgSize}

BOOTFS=${imagesDir}/${bootPartOutput}.vfat


# Some empty partitions are required ...
DONOTPOPULATE=${imagesDir}/nop.ext4
touch ${DONOTPOPULATE}

# get bootloader and format it with the right format

# Make a link for the right u-boot to use ?
UBOOTFILE=${U-BOOT}
#UBOOTPROJECT=$(ls ${WORKSPACE_LOC} | grep 'u-boot.*')
#NUBOOTFILE=$(ls ${UBOOTPROJECT} | grep 'u-boot.*\.bin$')

#if [ ! -z ${NUBOOTFILE} ] ; then
#	UBOOTFILE=${NUBOOTFILE}
#fi

echo "u-boot project : ${UBOOTPROJECT}"
echo "boot file : ${UBOOTFILE}"


# check rootfs existence
if [ ! -f ${ROOTFS} ] ; then
	echo "ERROR cannot retrieve ${ROOTFS}"
fi  

# call sdcard builder and generate the sdcard file

echo "==== Info ===="
echo " uboot-spl ${UBOOTSPL}"
echo " uboot ${UBOOT}"
echo " bootfs ${BOOTFS}"
echo " rootfs ${ROOTFS}"
echo " layout ${LAYOUT}"

do_create_sdcard_file ${LAYOUT} 


# remove the temp directory
rm -rf temp


do_buildstep_append

